"""
Package for training provably robust boosted decision tree ensembles.
Most code comes from the GitHub repository 'max-andr/provably-robust-boosting' by Maksym Andriushchenko and Matthias Hein.
"""
